package transfer

import (
	"context"
	"fmt"
	"log"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"github.com/y7ut/logtransfer/kafka"
)

var (
	Start          = make(chan *kafka.Customer)
	Close          = make(chan string)
	CustomerManger = make(map[string]*kafka.Customer, 0)
	MaxRetryTime   = 3
	mu             sync.Mutex
	closeWg        sync.WaitGroup
)

// type Kernel struct{
// 	CustomerManger map[string]*kafka.Customer
// 	mu 		       sync.Mutex
// }

func registerManger(c *kafka.Customer) {
	mu.Lock()
	CustomerManger[c.Reader.Config().Topic] = c
	mu.Unlock()
}

func getCustomer(topic string) (customer *kafka.Customer, ok bool) {
	mu.Lock()
	customer, ok = CustomerManger[topic]
	mu.Unlock()

	return customer, ok
}

func uninstallCustomer(topic string) {
	mu.Lock()
	delete(CustomerManger, topic)
	mu.Unlock()
}

func getRegisterTopics() (topics []string) {
	mu.Lock()
	for topic := range CustomerManger {
		topics = append(topics, topic)
	}
	mu.Unlock()
	return topics
}

// 核心启动
func Run() {
	go func() {
		for{
			select {
			case customer := <-Start:
				log.Println("Starting Customer...")
				registerManger(customer)
				go ReadingMessage(customer)

			case closer := <-Close:
				_, ok := getCustomer(closer)
				if !ok {
					log.Printf(" Customer %s unstall Failed ", closer)
					
				}
				// log.Printf(" Customer %s exit with offset(%d)", closer, c.Reader.Offset())
				closeWg.Done()

			}
		}
	}()

	for topic := range ChooseTopic() {
		GroupID := topic + "_group"
		currentCustomer := kafka.InitReader(topic, GroupID)
		log.Printf("Init Customer of (%s) success!", topic)
		Start <- currentCustomer
	}

	for sign := range sign() {
		switch sign {
		case syscall.SIGHUP, syscall.SIGINT, syscall.SIGTERM:
			log.Println("Safe Exit with:", sign)

			currentTopics := getRegisterTopics()
			closeWg.Add(len(currentTopics))
			for _, topic := range currentTopics {
				log.Printf(" Customer %s unstalling...", topic)
				Close <- topic
			}
			closeWg.Wait()
			log.Printf(" Success unstall %d Transfer", len(currentTopics))
			os.Exit(0)
		}
	}
}

func sign() <-chan os.Signal {
	c := make(chan os.Signal)
	// 监听信号
	signal.Notify(c, syscall.SIGHUP, syscall.SIGINT, syscall.SIGTERM, syscall.SIGQUIT, syscall.SIGUSR1, syscall.SIGUSR2)
	return c
}

// 從Kafka中消費消息，注意这里会提交commit offset
func ReadingMessage(c *kafka.Customer) {
	defer c.Reader.Close()

	log.Println("Start Customer success!")

	var trycount int
	for {
		select {
		case <-c.Ctx.Done():
			closeWg.Add(1)
			Close<- c.Reader.Config().Topic
				closeWg.Wait()
			log.Println("Close customer of Topic :", c.Reader.Config().Topic)
			return
		default:
			// 使用超时上下文
			timeoutCtx, cancel := context.WithTimeout(c.Ctx, 5*time.Second)
			m, err := c.Reader.FetchMessage(timeoutCtx)
			cancel()
			if err != nil {
				log.Println("failed to Reader messages:", err)
				trycount++
				if trycount >= MaxRetryTime {
					log.Println("Too many retries, Customer has Suspended  : ", err)
					c.Cancel()
				}
				continue
			}
			// 临时打印
			fmt.Printf("message at topic/partition/offset %v/%v/%v: %s = %s\n", m.Topic, m.Partition, m.Offset, string(m.Key), string(m.Value))

			// 使用超时上下文
			nextTimeoutCtx, nextCancel := context.WithTimeout(c.Ctx, 5*time.Second)
			err = c.Reader.CommitMessages(nextTimeoutCtx, m)
			nextCancel()
			if err != nil {
				cancel()
				log.Println("failed to commit messages:", err)
				trycount++
				if trycount >= MaxRetryTime {
					log.Println("Too many retries, Customer has Suspended  : ", err)
					c.Cancel()
				}
				continue
			}

			// 手动限速
			time.Sleep(1 * time.Second)
		}
	}
}
